"""
Tafra: a minimalist dataframe

Copyright (c) 2020 Derrick W. Turk and David S. Fulford

Author
------
Derrick W. Turk
David S. Fulford

Notes
-----
Created on April 25, 2020
"""
__all__ = ['GroupBy', 'Transform', 'IterateBy', 'InnerJoin', 'LeftJoin',
           'percentile', 'geomean', 'harmean']

import operator
from itertools import chain
import dataclasses as dc

import numpy as np

from typing import (Any, Callable, Dict, Mapping, List, Tuple, Optional, Union as _Union, Sequence,
                    Iterable, Iterator)
from typing import cast

try:
    from ._accel import (  # type: ignore[import-not-found]
        groupby_sum as _c_sum,
        groupby_mean as _c_mean,
        groupby_var as _c_var,
        groupby_min as _c_min,
        groupby_max as _c_max,
        groupby_count as _c_count,
        inner_join as _c_inner_join,
        left_join as _c_left_join,
    )
    _HAS_ACCEL = True
except ImportError:
    _HAS_ACCEL = False


# Vectorized aggregation functions that can bypass per-group Python loops.
# Maps function identity to a callable(data, labels, n_groups) -> result_array.
_VECTORIZED_AGGS: Dict[int, Callable[..., np.ndarray[Any, Any]]] = {}


def _sorted_segments(
    data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any], n: int
) -> Tuple[np.ndarray[Any, Any], np.ndarray[Any, Any], np.ndarray[Any, Any]]:
    """Sort data by group labels, return (sorted_data, offsets, counts)."""
    order = np.argsort(labels, kind='stable')
    sorted_data = data[order]
    counts = np.bincount(labels, minlength=n)
    offsets = np.zeros(n + 1, dtype=np.intp)
    np.cumsum(counts, out=offsets[1:])
    return sorted_data, offsets, counts


def _vec_sum(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
             n: int) -> np.ndarray[Any, Any]:
    return np.bincount(labels, weights=data.astype(float), minlength=n)


def _vec_mean(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
              n: int) -> np.ndarray[Any, Any]:
    sums = np.bincount(labels, weights=data.astype(float), minlength=n)
    counts = np.bincount(labels, minlength=n)
    return sums / counts


def _vec_var(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
             n: int) -> np.ndarray[Any, Any]:
    fdata = data.astype(float)
    counts = np.bincount(labels, minlength=n).astype(float)
    sums = np.bincount(labels, weights=fdata, minlength=n)
    sum_sq = np.bincount(labels, weights=fdata * fdata, minlength=n)
    mean = sums / counts
    return sum_sq / counts - mean * mean


def _vec_std(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
             n: int) -> np.ndarray[Any, Any]:
    return np.sqrt(_vec_var(data, labels, n))


def _vec_min(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
             n: int) -> np.ndarray[Any, Any]:
    sorted_data, offsets, _ = _sorted_segments(data, labels, n)
    return np.minimum.reduceat(sorted_data, offsets[:n])


def _vec_max(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
             n: int) -> np.ndarray[Any, Any]:
    sorted_data, offsets, _ = _sorted_segments(data, labels, n)
    return np.maximum.reduceat(sorted_data, offsets[:n])


def _vec_ptp(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
             n: int) -> np.ndarray[Any, Any]:
    sorted_data, offsets, _ = _sorted_segments(data, labels, n)
    return (np.maximum.reduceat(sorted_data, offsets[:n])
            - np.minimum.reduceat(sorted_data, offsets[:n]))


def _vec_prod(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
              n: int) -> np.ndarray[Any, Any]:
    sorted_data, offsets, _ = _sorted_segments(data, labels, n)
    return np.multiply.reduceat(sorted_data.astype(float), offsets[:n])


def _vec_any(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
             n: int) -> np.ndarray[Any, Any]:
    sorted_data, offsets, _ = _sorted_segments(data.astype(bool), labels, n)
    return np.logical_or.reduceat(sorted_data, offsets[:n])


def _vec_all(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
             n: int) -> np.ndarray[Any, Any]:
    sorted_data, offsets, _ = _sorted_segments(data.astype(bool), labels, n)
    return np.logical_and.reduceat(sorted_data, offsets[:n])


def _vec_len(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
             n: int) -> np.ndarray[Any, Any]:
    return np.bincount(labels, minlength=n)


def _vec_count_nonzero(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
                       n: int) -> np.ndarray[Any, Any]:
    return np.bincount(labels, weights=(data != 0).astype(float), minlength=n)


def _vec_median(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
                n: int) -> np.ndarray[Any, Any]:
    sorted_data, offsets, counts = _sorted_segments(data, labels, n)
    result = np.empty(n, dtype=float)
    for g in range(n):
        lo = offsets[g]
        hi = offsets[g + 1]
        seg = np.sort(sorted_data[lo:hi])
        c = int(counts[g])
        mid = c // 2
        if c % 2 == 1:
            result[g] = seg[mid]
        else:
            result[g] = (seg[mid - 1] + seg[mid]) / 2.0
    return result


def _vec_geomean(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
                 n: int) -> np.ndarray[Any, Any]:
    log_sums = np.bincount(labels, weights=np.log(data.astype(float)), minlength=n)
    counts = np.bincount(labels, minlength=n).astype(float)
    return np.exp(log_sums / counts)


def _vec_harmean(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
                 n: int) -> np.ndarray[Any, Any]:
    recip_sums = np.bincount(labels, weights=1.0 / data.astype(float), minlength=n)
    counts = np.bincount(labels, minlength=n).astype(float)
    return counts / recip_sums


def _make_vec_percentile(
    q: float
) -> Callable[[np.ndarray[Any, Any], np.ndarray[Any, Any], int], np.ndarray[Any, Any]]:
    """Create a vectorized percentile function for a given quantile."""
    def _vec_pct(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
                 n: int) -> np.ndarray[Any, Any]:
        sorted_data, offsets, counts = _sorted_segments(data, labels, n)
        result = np.empty(n, dtype=float)
        for g in range(n):
            lo = offsets[g]
            hi = offsets[g + 1]
            result[g] = np.percentile(sorted_data[lo:hi], q)
        return result
    return _vec_pct


class _PercentileAgg:
    """Callable aggregation for percentile(q). Registered for vectorized fast path."""
    __slots__ = ('q', '_vec_fn')

    def __init__(self, q: float) -> None:
        self.q = q
        self._vec_fn = _make_vec_percentile(q)
        _VECTORIZED_AGGS[id(self)] = self._vec_fn

    def __call__(self, data: np.ndarray[Any, Any]) -> Any:
        return np.percentile(data, self.q)

    def __repr__(self) -> str:
        return f'percentile({self.q})'


def percentile(q: float) -> _PercentileAgg:
    """
    Create a percentile aggregation function for use in ``group_by``.

    Parameters
    ----------
        q: float
            Percentile in range [0, 100].

    Returns
    -------
        agg: callable
            A callable suitable for ``group_by`` aggregation that also
            hits the vectorized fast path.

    Example
    -------
        >>> tf.group_by(['g'], {'p90': (percentile(90), 'value')})
    """
    return _PercentileAgg(q)


def geomean(data: np.ndarray[Any, Any]) -> Any:
    """Geometric mean aggregation for use in ``group_by``."""
    return np.exp(np.mean(np.log(data.astype(float))))


def harmean(data: np.ndarray[Any, Any]) -> Any:
    """Harmonic mean aggregation for use in ``group_by``."""
    return len(data) / np.sum(1.0 / data.astype(float))


def _register_vectorized() -> None:
    """Register known numpy reducers for vectorized GroupBy."""
    # Use C accelerated versions when available, fall back to numpy
    if _HAS_ACCEL:
        def _wrap_c(c_fn: Callable[..., Any]) -> Callable[..., np.ndarray[Any, Any]]:
            def _wrapped(data: np.ndarray[Any, Any], labels: np.ndarray[Any, Any],
                         n: int) -> np.ndarray[Any, Any]:
                return c_fn(labels.astype(np.int64, copy=False),
                            data.astype(np.float64, copy=False), n)
            return _wrapped
        _fast_sum = _wrap_c(_c_sum)
        _fast_mean = _wrap_c(_c_mean)
        _fast_var = _wrap_c(_c_var)
        _fast_min = _wrap_c(_c_min)
        _fast_max = _wrap_c(_c_max)
    else:
        _fast_sum = _vec_sum
        _fast_mean = _vec_mean
        _fast_var = _vec_var
        _fast_min = _vec_min
        _fast_max = _vec_max

    _fast_std = (lambda d, l, n: np.sqrt(_fast_var(d, l, n))) if _HAS_ACCEL else _vec_std

    _VECTORIZED_AGGS[id(np.sum)] = _fast_sum
    _VECTORIZED_AGGS[id(np.nansum)] = _fast_sum
    _VECTORIZED_AGGS[id(np.mean)] = _fast_mean
    _VECTORIZED_AGGS[id(np.nanmean)] = _fast_mean
    _VECTORIZED_AGGS[id(np.std)] = _fast_std
    _VECTORIZED_AGGS[id(np.nanstd)] = _fast_std
    _VECTORIZED_AGGS[id(np.var)] = _fast_var
    _VECTORIZED_AGGS[id(np.nanvar)] = _fast_var
    _VECTORIZED_AGGS[id(np.min)] = _fast_min
    _VECTORIZED_AGGS[id(np.nanmin)] = _fast_min
    _VECTORIZED_AGGS[id(np.amin)] = _fast_min
    _VECTORIZED_AGGS[id(np.max)] = _fast_max
    _VECTORIZED_AGGS[id(np.nanmax)] = _fast_max
    _VECTORIZED_AGGS[id(np.amax)] = _fast_max
    _VECTORIZED_AGGS[id(np.ptp)] = _vec_ptp
    _VECTORIZED_AGGS[id(np.prod)] = _vec_prod
    _VECTORIZED_AGGS[id(np.nanprod)] = _vec_prod
    _VECTORIZED_AGGS[id(np.any)] = _vec_any
    _VECTORIZED_AGGS[id(np.all)] = _vec_all
    _VECTORIZED_AGGS[id(np.median)] = _vec_median
    _VECTORIZED_AGGS[id(np.nanmedian)] = _vec_median
    _VECTORIZED_AGGS[id(np.count_nonzero)] = _vec_count_nonzero
    _VECTORIZED_AGGS[id(len)] = _vec_len
    _VECTORIZED_AGGS[id(sum)] = _vec_sum
    _VECTORIZED_AGGS[id(geomean)] = _vec_geomean
    _VECTORIZED_AGGS[id(harmean)] = _vec_harmean


_register_vectorized()


JOIN_OPS: Dict[str, Callable[[Any, Any], Any]] = {
    '==': operator.eq,
    '!=': operator.ne,
    '<': operator.lt,
    '<=': operator.le,
    '>': operator.gt,
    '>=': operator.ge
}

# for the passed argument to an aggregation
InitAggregation = Mapping[
    str,
    _Union[
        Callable[[np.ndarray[Any, Any]], Any],
        Tuple[Callable[[np.ndarray[Any, Any]], Any], str]
    ]
]


# for the result type of IterateBy
GroupDescription = Tuple[
    Tuple[Any, ...],  # tuple of unique values from group-by columns
    np.ndarray[Any, Any],  # int array of row indices into original tafra for this group
    'Tafra'  # sub-tafra for the group
]


class Union:
    """
    Union two :class:`Tafra` together. Analogy to SQL UNION or
    `pandas.append`. All column names and dtypes must match.
    """
    @staticmethod
    def _validate(left: 'Tafra', right: 'Tafra') -> None:
        """
        Validate the :class:`Tafra` before applying.
        """
        # These should be unreachable unless attributes were directly modified
        if len(left._data) != len(left._dtypes):
            raise ValueError('This `Tafra` length of data and dtypes do not match')
        if len(right._data) != len(right._dtypes):
            raise ValueError('right `Tafra` length of data and dtypes do not match')

        # ensure same number of columns
        if len(left._data) != len(right._data) or len(left._dtypes) != len(right._dtypes):
            raise ValueError(
                'This `Tafra` column count does not match right `Tafra` column count.')

        # ensure all columns in this `Tafra` exist in right `Tafra`
        # if len() is same AND all columns in this exist in right,
        # do not need to check right `Tafra` columns in this `Tafra`.
        for (data_column, value), (dtype_column, dtype) \
                in zip(left._data.items(), left._dtypes.items()):

            if data_column not in right._data or dtype_column not in right._dtypes:
                raise TypeError(
                    f'This `Tafra` column `{data_column}` does not exist in right `Tafra`.')

            elif value.dtype != right._data[data_column].dtype:
                raise TypeError(
                    f'This `Tafra` column `{data_column}` dtype `{value.dtype}` '
                    f'does not match right `Tafra` dtype `{right._data[data_column].dtype}`.')

            # should not happen unless dtypes manually changed, but let's check it
            elif dtype != right._dtypes[dtype_column]:
                raise TypeError(
                    f'This `Tafra` column `{data_column}` dtype `{dtype}` '
                    f'does not match right `Tafra` dtype `{right._dtypes[dtype_column]}`.')

    def apply(self, left: 'Tafra', right: 'Tafra') -> 'Tafra':
        """
        Apply the :class:`Union_` to the :class:`Tafra`.

        Parameters
        ----------
            left: Tafra
                The left :class:`Tafra` to union.

            right: Tafra
                The right :class:`Tafra` to union.

        Returns
        -------
            tafra: Tafra
                The unioned :class`Tafra`.
        """
        self._validate(left, right)

        return Tafra(
            {column: np.append(value, right._data[column]) for column, value in left._data.items()},
            left._dtypes.copy()
        )

    def apply_inplace(self, left: 'Tafra', right: 'Tafra') -> None:
        """
        In-place version.

        Apply the :class:`Union_` to the :class:`Tafra`.

        Parameters
        ----------
            left: Tafra
                The left :class:`Tafra` to union.

            right: Tafra
                The right :class:`Tafra` to union.

        Returns
        -------
            tafra: Tafra
                The unioned :class`Tafra`.
        """
        self._validate(left, right)

        for column, value in left._data.items():
            left._data[column] = np.append(value, right._data[column])
        left._update_rows()

@dc.dataclass
class GroupSet:
    """
    A `GroupSet` is the set of columns by which we construct our groups.
    """

    @staticmethod
    def _encode_columns(
        col_arrays: List[np.ndarray[Any, Any]]
    ) -> Tuple[List[np.ndarray[Any, Any]], List[Optional[np.ndarray[Any, Any]]]]:
        """
        Encode columns for structured array compatibility.
        StringDType columns are converted to integer codes via np.unique.

        Returns (encoded_arrays, codebooks) where codebooks[i] is the
        unique values array for encoded columns, or None if no encoding
        was needed.
        """
        encoded = []
        codebooks: List[Optional[np.ndarray[Any, Any]]] = []
        for c in col_arrays:
            if c.dtype.kind in ('T', 'U', 'S', 'O'):
                uniq, codes = np.unique(c, return_inverse=True)
                encoded.append(codes)
                codebooks.append(uniq)
            else:
                encoded.append(c)
                codebooks.append(None)
        return encoded, codebooks

    @staticmethod
    def _encode_columns_paired(
        left_cols: List[np.ndarray[Any, Any]],
        right_cols: List[np.ndarray[Any, Any]],
    ) -> Tuple[List[np.ndarray[Any, Any]], List[np.ndarray[Any, Any]]]:
        """
        Encode left and right column pairs with a shared codebook.
        Ensures the same string value gets the same integer code on both sides.
        """
        left_enc = []
        right_enc = []
        for lc, rc in zip(left_cols, right_cols):
            if lc.dtype.kind in ('T', 'U', 'S', 'O'):
                combined = np.concatenate([lc, rc])
                _, codes = np.unique(combined, return_inverse=True)
                left_enc.append(codes[:len(lc)])
                right_enc.append(codes[len(lc):])
            else:
                left_enc.append(lc)
                right_enc.append(rc)
        return left_enc, right_enc

    @staticmethod
    def _build_composite_key(
        encoded: List[np.ndarray[Any, Any]]
    ) -> np.ndarray[Any, Any]:
        """
        Combine multiple integer-coded columns into a single flat key.
        Uses positional encoding: key = c0 * N1*N2*... + c1 * N2*... + c2 * ...
        Falls back to structured array if values would overflow int64.
        """
        if len(encoded) == 1:
            return encoded[0]

        # compute cardinality of each column
        cards = [int(c.max()) + 1 for c in encoded]

        # check for overflow: product of all cardinalities must fit int64
        product = 1
        overflow = False
        for card in cards:
            product *= card
            if product > 2**62:
                overflow = True
                break

        if not overflow:
            # flat integer key via positional encoding
            key = np.zeros(len(encoded[0]), dtype=np.int64)
            multiplier = 1
            for c, card in zip(reversed(encoded), reversed(cards)):
                key += c.astype(np.int64) * multiplier
                multiplier *= card
            return key
        else:
            # fallback: structured array (slow but handles any cardinality)
            n_rows = len(encoded[0])
            dt = np.dtype([(f'f{i}', c.dtype) for i, c in enumerate(encoded)])
            data = np.empty(n_rows, dtype=dt)
            for i, c in enumerate(encoded):
                data[f'f{i}'] = c
            return data

    @staticmethod
    def _decode_unique(
        first_seen_indices: np.ndarray[Any, Any],
        col_arrays: List[np.ndarray[Any, Any]],
    ) -> List[Tuple[Any, ...]]:
        """Build unique group tuples by indexing original columns at first-seen rows."""
        if len(col_arrays) == 1:
            return [(v,) for v in col_arrays[0][first_seen_indices]]
        else:
            return [
                tuple(c[idx] for c in col_arrays)
                for idx in first_seen_indices
            ]

    @staticmethod
    def _direct_labels_firstseen(
        data: np.ndarray[Any, Any], n_rows: int
    ) -> Tuple[np.ndarray[Any, Any], np.ndarray[Any, Any], int]:
        """
        Assign first-seen-order group labels using direct array mapping.
        O(n + max_key) — no sort. Requires non-negative integer keys.

        Returns (labels, first_seen_indices, n_groups).
        """
        max_key = int(data.max())
        # sorted label map
        seen = np.zeros(max_key + 1, dtype=bool)
        seen[data] = True
        n_groups = int(seen.sum())
        sorted_map = np.empty(max_key + 1, dtype=np.intp)
        sorted_map[seen] = np.arange(n_groups)
        sorted_labels = sorted_map[data]
        # reorder to first-seen
        first_pos = np.empty(n_groups, dtype=np.intp)
        first_pos[sorted_labels[::-1]] = np.arange(n_rows - 1, -1, -1)
        order = np.argsort(first_pos)
        rank = np.empty(n_groups, dtype=np.intp)
        rank[order] = np.arange(n_groups, dtype=np.intp)
        labels = rank[sorted_labels]
        return labels, first_pos[order], n_groups

    @staticmethod
    def _direct_labels_sorted(
        data: np.ndarray[Any, Any]
    ) -> Tuple[np.ndarray[Any, Any], int]:
        """
        Assign sorted-order group labels using direct array mapping.
        O(n + max_key) — no sort. For Transform (order doesn't matter).

        Returns (labels, n_groups).
        """
        max_key = int(data.max())
        seen = np.zeros(max_key + 1, dtype=bool)
        seen[data] = True
        n_groups = int(seen.sum())
        label_map = np.empty(max_key + 1, dtype=np.intp)
        label_map[seen] = np.arange(n_groups, dtype=np.intp)
        return label_map[data], n_groups

    @staticmethod
    def _prepare_keys(
        tafra: 'Tafra', columns: Iterable[str]
    ) -> Tuple[np.ndarray[Any, Any], List[np.ndarray[Any, Any]]]:
        """Encode columns and build composite integer key."""
        cols = list(columns)
        col_arrays = [tafra._data[c] for c in cols]
        encoded, _ = GroupSet._encode_columns(col_arrays)
        if len(cols) == 1:
            data = encoded[0]
        else:
            data = GroupSet._build_composite_key(encoded)
        return data, col_arrays

    @staticmethod
    def _build_group_indices(
        tafra: 'Tafra', columns: Iterable[str]
    ) -> Tuple[List[Tuple[Any, ...]], List[np.ndarray[Any, Any]]]:
        """
        Build per-group row index arrays in a single pass.

        Returns (unique, group_indices) where:
        - unique: list of tuples of group-by values, in first-seen order
        - group_indices: list of int arrays (row positions for each group)
        """
        data, col_arrays = GroupSet._prepare_keys(tafra, columns)

        labels, first_seen_idx, n_groups = GroupSet._direct_labels_firstseen(
            data, tafra._rows)

        sorted_row_indices = np.argsort(labels, kind='stable')
        counts = np.bincount(labels, minlength=n_groups)
        splits = np.cumsum(counts[:-1])
        group_indices: List[np.ndarray[Any, Any]] = list(
            np.split(sorted_row_indices, splits))

        unique: List[Tuple[Any, ...]] = GroupSet._decode_unique(
            first_seen_idx, col_arrays)

        return unique, group_indices

    @staticmethod
    def _build_group_labels(
        tafra: 'Tafra', columns: Iterable[str]
    ) -> Tuple[List[Tuple[Any, ...]], np.ndarray[Any, Any], int]:
        """
        Build per-row group labels (integers 0..n_groups-1) in first-seen order.

        Returns (unique, labels, n_groups) where labels[i] is the group index
        for row i.
        """
        data, col_arrays = GroupSet._prepare_keys(tafra, columns)

        labels, first_seen_idx, n_groups = GroupSet._direct_labels_firstseen(
            data, tafra._rows)

        unique: List[Tuple[Any, ...]] = GroupSet._decode_unique(
            first_seen_idx, col_arrays)

        return unique, labels, n_groups

    @staticmethod
    def _validate(tafra: 'Tafra', columns: Iterable[str]) -> None:  # pragma: no cover
        """
        Validate the :class:`Tafra` before applying.
        """
        if tafra._rows < 1:
            raise ValueError('No rows exist in `tafra`.')
        tafra._validate_columns(columns)


@dc.dataclass
class AggMethod(GroupSet):
    """
    Basic methods for aggregations over a data table.
    """
    group_by_cols: Iterable[str]
    aggregation: dc.InitVar[InitAggregation]
    _aggregation: Mapping[
        str, Tuple[Callable[[np.ndarray[Any, Any]], Any], str]
    ] = dc.field(init=False)
    iter_fn: Mapping[str, Callable[[np.ndarray[Any, Any]], Any]]

    def __post_init__(self, aggregation: InitAggregation) -> None:
        self._aggregation = dict()
        for rename, agg in aggregation.items():
            if callable(agg):
                self._aggregation[rename] = (agg, rename)
            elif (isinstance(agg, Sequence) and len(agg) == 2
                  and callable(agg[0])):
                self._aggregation[rename] = agg
            else:
                raise ValueError(f'{rename}: {agg} is not a valid aggregation argument')

        for rename, agg in self.iter_fn.items():
            if not callable(agg):
                raise ValueError(f'{rename}: {agg} is not a valid aggregation argument')

    def result_factory(
        self, fn: Callable[[str, str], np.ndarray[Any, Any]]
    ) -> Dict[str, np.ndarray[Any, Any]]:
        """
        Factory function to generate the dict for the results set.
        A function to take the new column name and source column name
        and return an empty `np.ndarray` should be given.
        """
        return {
            rename: fn(rename, col) for rename, col in (
                *((col, col) for col in self.group_by_cols),
                *((rename, agg[1]) for rename, agg in self._aggregation.items())
            )
        }

    def iter_fn_factory(
        self, fn: Callable[[], np.ndarray[Any, Any]]
    ) -> Dict[str, np.ndarray[Any, Any]]:
        return {rename: fn() for rename in self.iter_fn.keys()}

    def apply(self, tafra: 'Tafra') -> 'Tafra':
        raise NotImplementedError


class GroupBy(AggMethod):
    """
    Aggregation by a set of unique values.

    Analogy to SQL ``GROUP BY``, not :meth:`pandas.DataFrame.groupby()`.

    Parameters
    ----------
        columns: Iterable[str]
            The column names to group by.

        aggregation: Mapping[str, Union[Callable[[np.ndarray], Any], \
            Optional. Tuple[Callable[[np.ndarray], Any], str]]]
            A mapping for columns and aggregation functions. Should be
            given as {'column': fn} or {'new_column': (fn, 'column')}.

        iter_fn: Mapping[str, Callable[[np.ndarray], Any]]
            Optional. A mapping for new columns names to the function to apply to
            the enumeration. Should be given as {'new_column': fn}.
    """

    def apply(self, tafra: 'Tafra') -> 'Tafra':
        """
        Apply the :class:`GroupBy` to the :class:`Tafra`.

        Parameters
        ----------
            tafra: Tafra
                The tafra to apply the operation to.

        Returns
        -------
            tafra: Tafra
                The aggregated :class:`Tafra`.
        """
        self._validate(tafra, (
            *self.group_by_cols,
            *(col for (_, col) in self._aggregation.values())
        ))

        # check if all aggregations can be vectorized
        all_vectorized = (
            not self.iter_fn
            and all(id(fn) in _VECTORIZED_AGGS
                    for fn, _ in self._aggregation.values())
        )

        if all_vectorized and self._aggregation:
            # fast path: vectorized aggregation via labels, no per-group loop
            unique, labels, n_groups = self._build_group_labels(
                tafra, self.group_by_cols)

            result: Dict[str, np.ndarray[Any, Any]] = {}
            for i, col in enumerate(self.group_by_cols):
                vals = tafra._data[col]
                # pick one representative value per group
                first_occurrence = np.empty(n_groups, dtype=np.intp)
                first_occurrence[labels] = np.arange(len(labels))
                # overwrite gives last, we want first — reverse
                first_occurrence[labels[::-1]] = np.arange(
                    len(labels) - 1, -1, -1)
                result[col] = vals[first_occurrence]

            for rename, (fn, col) in self._aggregation.items():
                vec_fn = _VECTORIZED_AGGS[id(fn)]
                result[rename] = vec_fn(tafra._data[col], labels, n_groups)

            return Tafra(result)

        # standard path: per-group loop
        unique, group_indices = self._build_group_indices(
            tafra, self.group_by_cols)
        n_groups = len(unique)

        result = self.result_factory(
            lambda rename, col: np.empty(n_groups, dtype=tafra._data[col].dtype))
        iter_fn = self.iter_fn_factory(lambda: np.ones(n_groups, dtype=int))

        for i, (u, rows) in enumerate(zip(unique, group_indices)):
            for val, col in zip(u, self.group_by_cols):
                result[col][i] = val

            for rename, (fn, col) in self._aggregation.items():
                result[rename][i] = fn(tafra._data[col][rows])

            for rename, fn in self.iter_fn.items():
                iter_fn[rename][i] = fn(np.full(len(rows), i, dtype=int))

        result.update(iter_fn)
        return Tafra(result)


class Transform(AggMethod):
    """
    Apply a function to each unique set of values and join to the original table.

    Analogy to :meth:`pandas.DataFrame.groupby().transform()`,
    i.e. a SQL ``GROUP BY`` and ``LEFT JOIN`` back to the original table.

    Parameters
    ----------
        group_by: Iterable[str]
            The column names to group by.

        aggregation: Mapping[str, Union[Callable[[np.ndarray], Any], \
        Tuple[Callable[[np.ndarray], Any], str]]]
            Optional. A mapping for columns and aggregation functions. Should be
            given as {'column': fn} or {'new_column': (fn, 'column')}.

        iter_fn: Mapping[str, Callable[[np.ndarray], Any]]
            Optional. A mapping for new columns names to the function to apply to
            the enumeration. Should be given as {'new_column': fn}.
    """

    def apply(self, tafra: 'Tafra') -> 'Tafra':
        """
        Apply the :class:`Transform` to the :class:`Tafra`.

        Parameters
        ----------
            tafra: Tafra
                The tafra to apply the operation to.

        Returns
        -------
            tafra: Tafra
                The transformed :class:`Tafra`.
        """
        self._validate(tafra, (
            *self.group_by_cols,
            *(col for (_, col) in self._aggregation.values())
        ))

        # check if all aggregations can be vectorized
        all_vectorized = (
            not self.iter_fn
            and all(id(fn) in _VECTORIZED_AGGS
                    for fn, _ in self._aggregation.values())
        )

        if all_vectorized and self._aggregation:
            # fast path: compute per-group aggregates, broadcast via labels
            # Transform doesn't need first-seen order — use direct mapping
            data, _ = self._prepare_keys(tafra, self.group_by_cols)
            labels, n_groups = self._direct_labels_sorted(data)

            result: Dict[str, np.ndarray[Any, Any]] = {}
            for col in self.group_by_cols:
                result[col] = tafra._data[col].copy()

            for rename, (fn, col) in self._aggregation.items():
                vec_fn = _VECTORIZED_AGGS[id(fn)]
                group_values = vec_fn(tafra._data[col], labels, n_groups)
                result[rename] = group_values[labels]

            return Tafra(result)

        # standard path: per-group loop
        unique, group_indices = self._build_group_indices(
            tafra, self.group_by_cols)

        result = self.result_factory(
            lambda rename, col: np.empty_like(tafra._data[col]))
        iter_fn = self.iter_fn_factory(lambda: np.ones(tafra._rows, dtype=int))

        for i, (u, rows) in enumerate(zip(unique, group_indices)):
            for col in self.group_by_cols:
                result[col][rows] = tafra._data[col][rows]

            for rename, (fn, col) in self._aggregation.items():
                result[rename][rows] = fn(tafra._data[col][rows])

            for rename, fn in self.iter_fn.items():
                iter_fn[rename][rows] = fn(np.full(len(rows), i, dtype=int))

        result.update(iter_fn)
        return Tafra(result)


@dc.dataclass
class IterateBy(GroupSet):
    """
    A generator that yields a :class:`Tafra` for each set of unique values.

    Analogy to `pandas.DataFrame.groupby()`, i.e. an Sequence of `Tafra` objects.
    Yields tuples of ((unique grouping values, ...), row indices array, subset tafra)

    Parameters
    ----------
        group_by: Iterable[str]
            The column names to group by.
    """
    group_by_cols: Iterable[str]

    def apply(self, tafra: 'Tafra') -> Iterator[GroupDescription]:
        """
        Apply the :class:`IterateBy` to the :class:`Tafra`.

        Parameters
        ----------
            tafra: Tafra
                The tafra to apply the operation to.

        Returns
        -------
            tafras: Iterator[GroupDescription]
                An iterator over the grouped :class:`Tafra`.
        """
        self._validate(tafra, self.group_by_cols)
        unique, group_indices = self._build_group_indices(tafra, self.group_by_cols)

        for u, rows in zip(unique, group_indices):
            yield (u, rows, tafra._ndindex(rows))


@dc.dataclass
class Join(GroupSet):
    """
    Base class for SQL-like JOINs.
    """
    on: Iterable[Tuple[str, str, str]]
    select: Iterable[str]

    def _validate_dtypes(self, l_table: 'Tafra', r_table: 'Tafra') -> None:
        for l_column, r_column, _ in self.on:
            l_value = l_table._data[l_column]
            r_value = r_table._data[r_column]
            l_dtype = l_table._dtypes[l_column]
            r_dtype = r_table._dtypes[r_column]

            if l_value.dtype != r_value.dtype:
                raise TypeError(
                    f'This `Tafra` column `{l_column}` dtype `{l_value.dtype}` '
                    f'does not match other `Tafra` dtype `{r_value.dtype}`.')

            # should not happen unless dtypes manually changed, but let's check it
            elif l_dtype != r_dtype:
                raise TypeError(
                    f'This `Tafra` column `{l_column}` dtype `{l_dtype}` '
                    f'does not match other `Tafra` dtype `{r_dtype}`.')

    @staticmethod
    def _validate_ops(ops: Iterable[str]) -> None:
        for op in ops:
            _op = JOIN_OPS.get(op, None)
            if _op is None:
                raise TypeError(f'The operator {op} is not valid.')

    @staticmethod
    def _build_composite_key(
        cols: List[np.ndarray[Any, Any]]
    ) -> np.ndarray[Any, Any]:
        """Build a single sortable key array from multiple columns."""
        if len(cols) == 1:
            return cols[0]
        # structured array for multi-column sort
        dt = np.dtype([(f'f{i}', c.dtype) for i, c in enumerate(cols)])
        key = np.empty(len(cols[0]), dtype=dt)
        for i, c in enumerate(cols):
            key[f'f{i}'] = c
        return key

    @staticmethod
    def _sort_merge_indices(
        left_key: np.ndarray[Any, Any],
        right_key: np.ndarray[Any, Any],
    ) -> Tuple[np.ndarray[Any, Any], np.ndarray[Any, Any]]:
        """
        Compute inner-join index pairs using sort-merge.
        Returns (left_indices, right_indices) as intp arrays.
        """
        right_order = np.argsort(right_key, kind='stable')
        right_sorted = right_key[right_order]

        left_lo = np.searchsorted(right_sorted, left_key, side='left')
        left_hi = np.searchsorted(right_sorted, left_key, side='right')
        counts = left_hi - left_lo

        total = int(counts.sum())
        if total == 0:
            return np.array([], dtype=np.intp), np.array([], dtype=np.intp)

        # left indices: repeat each left row index by its match count
        li = np.repeat(np.arange(len(left_key), dtype=np.intp), counts)

        # right indices: for each left row, enumerate its match range
        # build offsets into right_sorted for each output row
        offsets = np.repeat(left_lo, counts)
        group_starts = np.cumsum(counts) - counts
        within = np.arange(total, dtype=np.intp) - np.repeat(group_starts, counts)
        ri = right_order[offsets + within]

        return li, ri

    @staticmethod
    def _left_join_indices(
        left_key: np.ndarray[Any, Any],
        right_key: np.ndarray[Any, Any],
    ) -> Tuple[np.ndarray[Any, Any], np.ndarray[Any, Any], bool]:
        """
        Compute left-join index pairs using sort-merge.
        Unmatched left rows get right index = -1.
        Returns (left_indices, right_indices, has_null).
        """
        right_order = np.argsort(right_key, kind='stable')
        right_sorted = right_key[right_order]

        left_lo = np.searchsorted(right_sorted, left_key, side='left')
        left_hi = np.searchsorted(right_sorted, left_key, side='right')
        counts = left_hi - left_lo

        # unmatched left rows get count=0 → force to 1 (with sentinel)
        unmatched = counts == 0
        has_null = bool(np.any(unmatched))
        out_counts = np.where(unmatched, 1, counts)

        total = int(out_counts.sum())
        li = np.repeat(np.arange(len(left_key), dtype=np.intp), out_counts)

        # right indices
        ri = np.empty(total, dtype=np.intp)
        pos = 0
        if has_null:
            # mixed matched/unmatched — fill per group
            for i in range(len(left_key)):
                c = int(counts[i])
                if c > 0:
                    ri[pos:pos + c] = right_order[left_lo[i]:left_hi[i]]
                    pos += c
                else:
                    ri[pos] = -1
                    pos += 1
        else:
            # all matched — fully vectorized
            offsets = np.repeat(left_lo, counts)
            group_starts = np.cumsum(counts) - counts
            within = np.arange(total, dtype=np.intp) - np.repeat(
                group_starts, counts)
            ri = right_order[offsets + within]

        return li, ri, has_null

    def _resolve_join_cols(
        self, left_t: 'Tafra', right_t: 'Tafra'
    ) -> Tuple[List[str], Dict[str, str]]:
        """Compute deduplicated output columns and dtypes."""
        seen_cols: Dict[str, None] = {}
        for c in chain(left_t._data.keys(), right_t._data.keys()):
            if not self.select or c in self.select:
                seen_cols[c] = None
        join_cols = list(seen_cols.keys())
        dtypes: Dict[str, str] = {
            c: d for c, d in chain(
                right_t._dtypes.items(), left_t._dtypes.items())
            if c in seen_cols
        }
        return join_cols, dtypes

    def apply(self, left_t: 'Tafra', right_t: 'Tafra') -> 'Tafra':
        raise NotImplementedError


class InnerJoin(Join):
    """
    An inner join.

    Analogy to SQL INNER JOIN, or `pandas.merge(..., how='inner')`,

    Parameters
    ----------
        right: Tafra
            The right-side :class:`Tafra` to join.

        on: Iterable[Tuple[str, str, str]]
            The columns and operator to join on. Should be given as
            ('left column', 'right column', 'op') Valid ops are:

            '==' : equal to
            '!=' : not equal to
            '<'  : less than
            '<=' : less than or equal to
            '>'  : greater than
            '>=' : greater than or equal to

        select: Iterable[str] = []
            The columns to return. If not given, all unique columns names
            are returned. If the column exists in both :class`Tafra`,
            prefers the left over the right.
    """

    def apply(self, left_t: 'Tafra', right_t: 'Tafra') -> 'Tafra':
        """
        Apply the :class:`InnerJoin` to the :class:`Tafra`.

        Parameters
        ----------
            left_t: Tafra
                The left tafra to join.

            right_t: Tafra
                The right tafra to join.

        Returns
        -------
            tafra: Tafra
                The joined :class:`Tafra`.
        """
        left_cols, right_cols, ops = list(zip(*self.on))
        self._validate(left_t, left_cols)
        self._validate(right_t, right_cols)
        self._validate_dtypes(left_t, right_t)
        self._validate_ops(ops)

        join_cols, dtypes = self._resolve_join_cols(left_t, right_t)
        all_equi = all(op_str == '==' for _, _, op_str in self.on)

        if all_equi:
            # Encode left+right together for consistent codebooks
            left_cols_data = [left_t._data[lc] for lc, _, _ in self.on]
            right_cols_data = [right_t._data[rc] for _, rc, _ in self.on]
            l_enc, r_enc = GroupSet._encode_columns_paired(
                left_cols_data, right_cols_data)
            left_key = GroupSet._build_composite_key(l_enc)
            right_key = GroupSet._build_composite_key(r_enc)

            if _HAS_ACCEL:
                li, ri = _c_inner_join(
                    np.ascontiguousarray(left_key, dtype=np.int64),
                    np.ascontiguousarray(right_key, dtype=np.int64))
            else:
                li, ri = self._sort_merge_indices(left_key, right_key)

            if len(li) == 0:
                return Tafra(
                    {c: np.array(
                        [], dtype=left_t._data[c].dtype
                        if c in left_t._data
                        else right_t._data[c].dtype)
                     for c in join_cols},
                    dtypes
                )

            result: Dict[str, np.ndarray[Any, Any]] = {}
            for c in join_cols:
                if c in left_t._data:
                    result[c] = left_t._data[c][li]
                else:
                    result[c] = right_t._data[c][ri]

            return Tafra(result, dtypes)

        else:
            _on = tuple(
                (left_col, right_col, JOIN_OPS[op])
                for left_col, right_col, op in self.on
            )
            right_rows = np.empty(right_t._rows, dtype=bool)
            join: Dict[str, List[Any]] = {c: [] for c in join_cols}

            for i in range(left_t._rows):
                right_rows[:] = True
                for left_col, right_col, op in _on:
                    right_rows &= op(
                        left_t._data[left_col][i], right_t._data[right_col])

                right_count = int(np.sum(right_rows))
                if right_count <= 0:
                    continue

                for column in join_cols:
                    if column in left_t._data:
                        join[column].extend(
                            [left_t._data[column][i]] * right_count)
                    elif column in right_t._data:
                        join[column].extend(
                            right_t._data[column][right_rows])

            return Tafra(
                {c: np.array(v) for c, v in join.items()},
                dtypes
            )


class LeftJoin(Join):
    """
    A left join.

    Analogy to SQL LEFT JOIN, or `pandas.merge(..., how='left')`,

    Parameters
    ----------
        right: Tafra
            The right-side :class:`Tafra` to join.

        on: Iterable[Tuple[str, str, str]]
            The columns and operator to join on. Should be given as
            ('left column', 'right column', 'op') Valid ops are:

            '==' : equal to
            '!=' : not equal to
            '<'  : less than
            '<=' : less than or equal to
            '>'  : greater than
            '>=' : greater than or equal to

        select: Iterable[str] = []
            The columns to return. If not given, all unique columns names
            are returned. If the column exists in both :class`Tafra`,
            prefers the left over the right.
    """

    def apply(self, left_t: 'Tafra', right_t: 'Tafra') -> 'Tafra':
        """
        Apply the :class:`LeftJoin` to the :class:`Tafra`.

        Parameters
        ----------
            left_t: Tafra
                The left tafra to join.

            right_t: Tafra
                The right tafra to join.

        Returns
        -------
            tafra: Tafra
                The joined :class:`Tafra`.
        """
        left_cols, right_cols, ops = list(zip(*self.on))
        self._validate(left_t, left_cols)
        self._validate(right_t, right_cols)
        self._validate_dtypes(left_t, right_t)
        self._validate_ops(ops)

        join_cols, dtypes = self._resolve_join_cols(left_t, right_t)
        all_equi = all(op_str == '==' for _, _, op_str in self.on)

        if all_equi:
            left_cols_data = [left_t._data[lc] for lc, _, _ in self.on]
            right_cols_data = [right_t._data[rc] for _, rc, _ in self.on]
            l_enc, r_enc = GroupSet._encode_columns_paired(
                left_cols_data, right_cols_data)
            left_key = GroupSet._build_composite_key(l_enc)
            right_key = GroupSet._build_composite_key(r_enc)

            if _HAS_ACCEL:
                li, ri, has_null = _c_left_join(
                    np.ascontiguousarray(left_key, dtype=np.int64),
                    np.ascontiguousarray(right_key, dtype=np.int64))
            else:
                li, ri, has_null = self._left_join_indices(left_key, right_key)

            if has_null:
                for c in join_cols:
                    if c not in left_t._data and dtypes.get(c) != 'object':
                        dtypes[c] = 'object'

            result: Dict[str, np.ndarray[Any, Any]] = {}
            matched = ri >= 0
            for c in join_cols:
                if c in left_t._data:
                    result[c] = left_t._data[c][li]
                else:
                    # right column: fill matched rows, None for unmatched
                    if has_null:
                        out = np.empty(len(li), dtype=object)
                        out[matched] = right_t._data[c][ri[matched]]
                        out[~matched] = None
                        result[c] = out
                    else:
                        result[c] = right_t._data[c][ri]

            return Tafra(result, dtypes)

        else:
            _on = tuple(
                (left_col, right_col, JOIN_OPS[op])
                for left_col, right_col, op in self.on
            )
            right_rows = np.empty(right_t._rows, dtype=bool)
            join: Dict[str, List[Any]] = {c: [] for c in join_cols}

            for i in range(left_t._rows):
                right_rows[:] = True
                for left_col, right_col, op in _on:
                    right_rows &= op(
                        left_t._data[left_col][i], right_t._data[right_col])

                right_count = int(np.sum(right_rows))

                for column in join_cols:
                    if column in left_t._data:
                        join[column].extend(
                            [left_t._data[column][i]] * max(1, right_count))
                    elif column in right_t._data:
                        if right_count <= 0:
                            join[column].append(None)
                            if dtypes[column] != 'object':
                                dtypes[column] = 'object'
                        else:
                            join[column].extend(
                                right_t._data[column][right_rows])

            return Tafra(
                {c: np.array(v) for c, v in join.items()},
                dtypes
            )


@dc.dataclass
class CrossJoin(Join):
    """
    A cross join.

    Analogy to SQL CROSS JOIN, or `pandas.merge(..., how='outer')
    using temporary columns of static value to intersect all rows`.

    Parameters
    ----------
        right: Tafra
            The right-side :class:`Tafra` to join.

        select: Iterable[str] = []
            The columns to return. If not given, all unique columns names
            are returned. If the column exists in both :class`Tafra`,
            prefers the left over the right.
    """

    def apply(self, left_t: 'Tafra', right_t: 'Tafra') -> 'Tafra':
        """
        Apply the :class:`CrossJoin` to the :class:`Tafra`.

        Parameters
        ----------
            left_t: Tafra
                The left tafra to join.

            right_t: Tafra
                The right tafra to join.

        Returns
        -------
            tafra: Tafra
                The joined :class:`Tafra`.
        """
        self._validate_dtypes(left_t, right_t)

        left_rows = left_t._rows
        right_rows = right_t._rows

        select = set(self.select)
        if len(select) > 0:
            left_cols = list(select.intersection(list(left_t._data.keys())))
            right_cols = list(select.intersection(list(right_t._data.keys())))

            if len(left_cols) == 0:
                raise IndexError('No columns given to select from left `Tafra`.')
            if len(right_cols) == 0:
                raise IndexError('No columns given to select from right `Tafra`.')

        else:
            left_cols = list(left_t._data.keys())
            right_cols = list(right_t._data.keys())

        left_new = Tafra(left_t[left_cols].key_map(np.tile, reps=right_rows))
        right_new = Tafra(right_t[right_cols].key_map(np.tile, reps=left_rows))

        left_new.update_inplace(right_new)

        return left_new


# Import here to resolve circular dependency
from .base import Tafra



"""
Benchmarks for tafra performance.

Run with:
    python test/bench_tafra.py
    pytest test/bench_tafra.py -v  (if pytest-benchmark is installed)
"""
import time
import numpy as np
from contextlib import contextmanager
from typing import Callable, Any, Dict, List

from tafra import Tafra, GroupBy, Transform, IterateBy, InnerJoin, LeftJoin


@contextmanager
def timer(label: str):
    t0 = time.perf_counter()
    yield
    elapsed = time.perf_counter() - t0
    print(f"  {label}: {elapsed * 1000:.2f} ms")


def make_tafra(n_rows: int, n_groups: int) -> Tafra:
    rng = np.random.default_rng(42)
    return Tafra({
        'group_a': rng.integers(0, n_groups, size=n_rows).astype('int64'),
        'group_b': rng.choice(['x', 'y', 'z'], size=n_rows),
        'value1': rng.standard_normal(n_rows),
        'value2': rng.standard_normal(n_rows),
        'value3': rng.integers(0, 1000, size=n_rows).astype('float64'),
    })


def make_join_tafras(left_rows: int, right_rows: int) -> tuple:
    rng = np.random.default_rng(42)
    left = Tafra({
        'key': rng.integers(0, 100, size=left_rows).astype('int64'),
        'left_val': rng.standard_normal(left_rows),
    })
    right = Tafra({
        'key': rng.integers(0, 100, size=right_rows).astype('int64'),
        'right_val': rng.standard_normal(right_rows),
    })
    return left, right


def bench_construction(n_rows: int, n_iter: int = 100) -> None:
    rng = np.random.default_rng(42)
    data = {
        'a': rng.standard_normal(n_rows),
        'b': rng.standard_normal(n_rows),
        'c': rng.integers(0, 100, size=n_rows).astype('int64'),
    }

    with timer(f"Tafra() validated, {n_rows} rows x {n_iter} iters"):
        for _ in range(n_iter):
            Tafra(data.copy())

    with timer(f"Tafra() unvalidated, {n_rows} rows x {n_iter} iters"):
        for _ in range(n_iter):
            Tafra({k: v for k, v in data.items()}, validate=False)


def bench_getitem(n_rows: int, n_iter: int = 10000) -> None:
    t = make_tafra(n_rows, 50)

    with timer(f"__getitem__ str, {n_iter} iters"):
        for _ in range(n_iter):
            t['value1']

    with timer(f"__getitem__ list[str], {n_iter} iters"):
        for _ in range(n_iter):
            t[['value1', 'value2']]

    mask = t['group_a'] < 25
    with timer(f"__getitem__ bool mask, {n_iter} iters"):
        for _ in range(n_iter):
            t[mask]


def bench_setitem(n_rows: int, n_iter: int = 1000) -> None:
    t = make_tafra(n_rows, 50)
    new_values = np.ones(n_rows)

    with timer(f"__setitem__ existing col, {n_iter} iters"):
        for _ in range(n_iter):
            t['value1'] = new_values

    with timer(f"__setitem__ new col, {n_iter} iters"):
        for i in range(n_iter):
            t[f'new_{i}'] = new_values


def bench_groupby(n_rows: int, n_groups: int) -> None:
    t = make_tafra(n_rows, n_groups)

    gb = GroupBy(
        group_by_cols=['group_a'],
        aggregation={'mean_v1': (np.mean, 'value1'), 'sum_v2': (np.sum, 'value2')},
        iter_fn={},
    )
    with timer(f"GroupBy 1 col, {n_groups} groups, {n_rows} rows"):
        gb.apply(t)

    gb2 = GroupBy(
        group_by_cols=['group_a', 'group_b'],
        aggregation={'mean_v1': (np.mean, 'value1')},
        iter_fn={},
    )
    unique_combos = len(set(zip(t['group_a'], t['group_b'])))
    with timer(f"GroupBy 2 cols, ~{unique_combos} groups, {n_rows} rows"):
        gb2.apply(t)


def bench_transform(n_rows: int, n_groups: int) -> None:
    t = make_tafra(n_rows, n_groups)

    tr = Transform(
        group_by_cols=['group_a'],
        aggregation={'mean_v1': (np.mean, 'value1'), 'sum_v2': (np.sum, 'value2')},
        iter_fn={},
    )
    with timer(f"Transform 1 col, {n_groups} groups, {n_rows} rows"):
        tr.apply(t)


def bench_iterateby(n_rows: int, n_groups: int) -> None:
    t = make_tafra(n_rows, n_groups)

    ib = IterateBy(group_by_cols=['group_a'])
    with timer(f"IterateBy 1 col, {n_groups} groups, {n_rows} rows"):
        list(ib.apply(t))


def bench_innerjoin(left_rows: int, right_rows: int) -> None:
    left, right = make_join_tafras(left_rows, right_rows)

    ij = InnerJoin(
        on=[('key', 'key', '==')],
        select=[],
    )
    with timer(f"InnerJoin {left_rows}x{right_rows}"):
        ij.apply(left, right)


def bench_leftjoin(left_rows: int, right_rows: int) -> None:
    left, right = make_join_tafras(left_rows, right_rows)

    lj = LeftJoin(
        on=[('key', 'key', '==')],
        select=[],
    )
    with timer(f"LeftJoin {left_rows}x{right_rows}"):
        lj.apply(left, right)


def bench_build_group_indices(n_rows: int, n_groups: int) -> None:
    t = make_tafra(n_rows, n_groups)

    with timer(f"_build_group_indices 1 col, {n_rows} rows"):
        GroupBy._build_group_indices(t, ['group_a'])

    with timer(f"_build_group_indices 2 cols, {n_rows} rows"):
        GroupBy._build_group_indices(t, ['group_a', 'group_b'])


def bench_validate_columns(n_rows: int, n_iter: int = 10000) -> None:
    t = make_tafra(n_rows, 50)
    cols = list(t._data.keys())

    with timer(f"_validate_columns {len(cols)} cols x {n_iter} iters"):
        for _ in range(n_iter):
            t._validate_columns(cols)


def bench_ensure_valid(n_rows: int, n_iter: int = 1000) -> None:
    t = make_tafra(n_rows, 50)
    arr = np.ones(n_rows)

    with timer(f"_ensure_valid ndarray, {n_iter} iters"):
        for _ in range(n_iter):
            t._ensure_valid('value1', arr)

    with timer(f"_ensure_valid scalar, {n_iter} iters"):
        for _ in range(n_iter):
            t._ensure_valid('value1', 1.0)


def bench_copy(n_rows: int, n_iter: int = 100) -> None:
    t = make_tafra(n_rows, 50)

    with timer(f"copy(), {n_rows} rows x 5 cols x {n_iter} iters"):
        for _ in range(n_iter):
            t.copy()


def bench_select(n_rows: int, n_iter: int = 1000) -> None:
    t = make_tafra(n_rows, 50)

    with timer(f"select() 2 cols, {n_iter} iters"):
        for _ in range(n_iter):
            t.select(['value1', 'value2'])


def bench_format_dtype(n_iter: int = 10000) -> None:
    dt = np.dtype('float64')

    with timer(f"_format_dtype, {n_iter} iters"):
        for _ in range(n_iter):
            Tafra._format_dtype(dt)


def bench_parse_dtype(n_iter: int = 10000) -> None:
    from tafra import object_formatter
    arr = np.ones(100)

    with timer(f"parse_dtype (non-object), {n_iter} iters"):
        for _ in range(n_iter):
            object_formatter.parse_dtype(arr)


def _heavy_worker(sub: 'Tafra') -> 'Tafra':
    """Simulate expensive per-group work (~20-50ms): model fit + forecast."""
    v = sub['value1']
    t = np.linspace(0, 1, 2000)
    result = np.zeros_like(t)
    for _ in range(100):
        result += np.sin(np.mean(v) * t) * np.exp(-np.std(v) * t)
        result += np.cos(np.var(v) * t) * np.log1p(np.abs(t * np.mean(v)))
    return Tafra({
        'group_a': sub['group_a'][:1],
        'forecast': np.array([np.sum(result)]),
    }, validate=False)


def _light_worker(sub: 'Tafra') -> 'Tafra':
    """Light per-group work: simple aggregations."""
    return Tafra({
        'group_a': sub['group_a'][:1],
        'mean_v1': np.array([np.mean(sub['value1'])]),
        'std_v1': np.array([np.std(sub['value1'])]),
    }, validate=False)


def _warmup(x: int) -> int:
    return x


def bench_partition_vs_groupby(n_rows: int, n_groups: int,
                               pool: 'ProcessPoolExecutor',
                               n_workers: int) -> None:
    t = make_tafra(n_rows, n_groups)

    print(f"  [{n_groups} groups, {n_rows} rows, {n_workers} workers]")

    with timer(f"  group_by (light agg)"):
        t.group_by(['group_a'], {
            'mean_v1': (np.mean, 'value1'),
            'std_v1': (np.std, 'value1'),
        })

    with timer(f"  partition+serial (light)"):
        parts = t.partition(['group_a'])
        results = [_light_worker(sub) for _, sub in parts]
        Tafra.concat(results)

    with timer(f"  partition+serial (heavy)"):
        parts = t.partition(['group_a'])
        results = [_heavy_worker(sub) for _, sub in parts]
        Tafra.concat(results)

    with timer(f"  partition+{n_workers} workers (heavy)"):
        parts = t.partition(['group_a'])
        results = list(pool.map(_heavy_worker, [sub for _, sub in parts]))
        Tafra.concat(results)


if __name__ == '__main__':
    print("=" * 60)
    print("Tafra Performance Benchmarks")
    print("=" * 60)

    print("\n--- Construction ---")
    bench_construction(1000)
    bench_construction(100_000)

    print("\n--- Access ---")
    bench_getitem(10_000)
    bench_setitem(10_000)
    bench_select(10_000)

    print("\n--- Internal hot paths ---")
    bench_validate_columns(10_000)
    bench_ensure_valid(10_000)
    bench_format_dtype()
    bench_parse_dtype()
    bench_copy(10_000)

    print("\n--- Unique groups ---")
    bench_build_group_indices(10_000, 100)
    bench_build_group_indices(100_000, 1000)

    print("\n--- GroupBy ---")
    bench_groupby(10_000, 50)
    bench_groupby(10_000, 500)
    bench_groupby(100_000, 100)

    print("\n--- Transform ---")
    bench_transform(10_000, 50)
    bench_transform(10_000, 500)
    bench_transform(100_000, 100)

    print("\n--- IterateBy ---")
    bench_iterateby(10_000, 50)
    bench_iterateby(100_000, 100)

    print("\n--- Partition vs GroupBy ---")
    from concurrent.futures import ProcessPoolExecutor
    import os
    n_workers = min(os.cpu_count() or 4, 8)
    pool = ProcessPoolExecutor(max_workers=n_workers)
    list(pool.map(_warmup, range(n_workers)))  # warm up workers
    bench_partition_vs_groupby(10_000, 50, pool, n_workers)
    bench_partition_vs_groupby(100_000, 100, pool, n_workers)
    bench_partition_vs_groupby(100_000, 1000, pool, n_workers)
    pool.shutdown(wait=True)

    print("\n--- Joins ---")
    bench_innerjoin(500, 500)
    bench_innerjoin(1000, 1000)
    bench_leftjoin(500, 500)
    bench_leftjoin(1000, 1000)

    print("\n--- Joins (larger) ---")
    bench_innerjoin(5000, 5000)
    bench_innerjoin(10000, 10000)
    bench_leftjoin(5000, 5000)
    bench_leftjoin(10000, 10000)

    print("\n" + "=" * 60)
    print("Done.")



"""
Tafra vs pandas vs polars performance comparison.

Run with: python test/bench_vs_pandas_vs_polars.py
"""
import time
import math
import sys
import numpy as np
import pandas as pd

from tafra import Tafra

try:
    import polars as pl
    HAS_POLARS = True
except ImportError:
    HAS_POLARS = False

try:
    from tafra._accel import groupby_sum as _c_test
    HAS_ACCEL = True
except ImportError:
    HAS_ACCEL = False

import tafra.group as _grp


def median_of(fn, n=7):
    """Return median timing in ms."""
    times = []
    for _ in range(n):
        t0 = time.perf_counter()
        fn()
        times.append((time.perf_counter() - t0) * 1000)
    times.sort()
    return times[n // 2]


def _disable_accel():
    """Temporarily disable C acceleration."""
    saved = _grp._HAS_ACCEL
    _grp._HAS_ACCEL = False
    # re-register vectorized aggs with Python fallbacks
    _grp._VECTORIZED_AGGS.clear()
    _grp._register_vectorized()
    return saved


def _restore_accel(saved):
    """Restore C acceleration state."""
    _grp._HAS_ACCEL = saved
    _grp._VECTORIZED_AGGS.clear()
    _grp._register_vectorized()


def _tafra_time(fn):
    """Time fn with accel on and off, return (accel_ms, pure_ms)."""
    if HAS_ACCEL:
        t_accel = median_of(fn)
        saved = _disable_accel()
        t_pure = median_of(fn)
        _restore_accel(saved)
        return t_accel, t_pure
    else:
        t_pure = median_of(fn)
        return None, t_pure


def fmt(val, best=False):
    if val is None:
        return "n/a".rjust(9)
    s = f"{val:.2f}".rjust(9)
    if best:
        s = f"*{val:.2f}".rjust(9)
    return s


def print_row(label, t_accel, t_pure, t_pandas, t_polars):
    """Print one benchmark row, marking the best."""
    vals = {'accel': t_accel, 'pure': t_pure, 'pandas': t_pandas, 'polars': t_polars}
    real_vals = {k: v for k, v in vals.items() if v is not None}
    best_key = min(real_vals, key=real_vals.get) if real_vals else None

    print(f"  {label:<35s}"
          f" {fmt(t_accel, best_key == 'accel')}"
          f" {fmt(t_pure, best_key == 'pure')}"
          f" {fmt(t_pandas, best_key == 'pandas')}"
          f" {fmt(t_polars, best_key == 'polars')}")


def print_header():
    accel_label = "Tafra+C" if HAS_ACCEL else ""
    print(f"  {'Benchmark':<35s}"
          f" {accel_label:>9s}"
          f" {'Tafra':>9s}"
          f" {'pandas':>9s}"
          f" {'polars':>9s}")
    print(f"  {'-'*35} {'-'*9} {'-'*9} {'-'*9} {'-'*9}")


# ============================================================
# Benchmarks
# ============================================================

def bench_construction():
    print("\n--- Construction (100k rows, 5 cols) ---")
    print_header()
    rng = np.random.default_rng(42)
    data = {
        'a': rng.standard_normal(100_000),
        'b': rng.standard_normal(100_000),
        'c': rng.integers(0, 100, size=100_000).astype('int64'),
        'd': rng.standard_normal(100_000),
        'e': rng.standard_normal(100_000),
    }

    # construction doesn't use accel, same for both
    t_tafra = median_of(lambda: Tafra(data.copy()), n=21)
    t_pandas = median_of(lambda: pd.DataFrame(data.copy()), n=21)
    t_polars = median_of(lambda: pl.DataFrame(data.copy()), n=21) if HAS_POLARS else None
    print_row("Tafra(data)", t_tafra, t_tafra, t_pandas, t_polars)


def bench_column_access():
    print("\n--- Column Access (100k rows, 10k iters) ---")
    print_header()
    rng = np.random.default_rng(42)
    n = 100_000
    tf = Tafra({'x': rng.standard_normal(n), 'y': rng.standard_normal(n)})
    df = pd.DataFrame(tf.data)
    n_iter = 10_000

    t_tafra = median_of(lambda: [tf['x'] for _ in range(n_iter)])
    t_pandas = median_of(lambda: [df['x'] for _ in range(n_iter)])
    t_polars = None
    if HAS_POLARS:
        plf = pl.DataFrame(tf.data)
        t_polars = median_of(lambda: [plf['x'] for _ in range(n_iter)])
    print_row("tf['x'] x 10k", t_tafra, t_tafra, t_pandas, t_polars)


def bench_row_mapping():
    print("\n--- Row Mapping (100 wells x 101 time steps) ---")
    print_header()
    np.random.seed(42)
    tf = Tafra({
        'wellid': np.arange(0, 100),
        'qi': np.random.lognormal(np.log(2000.), np.log(3000. / 1000.) / (2 * 1.28), 100),
        'Di': np.random.uniform(.5, .9, 100),
        'bi': np.random.normal(1.0, .2, 100)
    })
    df = pd.DataFrame(tf.data)
    t = 10 ** np.linspace(0, 4, 101)

    def tan_to_nominal(D):
        return -math.log1p(-D)

    def sec_to_nominal(D, b):
        if b <= 1e-4:
            return tan_to_nominal(D)
        return ((1.0 - D) ** -b - 1.0) / b

    def hyp(qi, Di, bi, t):
        Dn = sec_to_nominal(Di, bi)
        if bi <= 1e-4:
            return qi * np.exp(-Dn * t)
        return qi / (1.0 + Dn * bi * t) ** (1.0 / bi)

    def tuple_mapper(tf_row):
        return tf_row.wellid, hyp(tf_row.qi, tf_row.Di, tf_row.bi, t)

    t_tafra = median_of(lambda: Tafra(tf.tuple_map(tuple_mapper)))
    t_pandas = median_of(
        lambda: pd.DataFrame(dict(tuple_mapper(row) for row in df.itertuples())))
    print_row("tuple_map", t_tafra, t_tafra, t_pandas, None)


def bench_groupby():
    print("\n--- GroupBy (sum + mean) ---")
    print_header()
    rng = np.random.default_rng(42)

    for n_rows, n_groups, label in [
        (10_000, 50, "10k rows, 50 grp"),
        (10_000, 500, "10k rows, 500 grp"),
        (100_000, 100, "100k rows, 100 grp"),
        (100_000, 1000, "100k rows, 1k grp"),
        (1_000_000, 100, "1M rows, 100 grp"),
        (1_000_000, 10000, "1M rows, 10k grp"),
    ]:
        tf = Tafra({
            'group': rng.integers(0, n_groups, size=n_rows).astype('int64'),
            'value': rng.standard_normal(n_rows),
        })
        df = pd.DataFrame(tf.data)

        t_accel, t_pure = _tafra_time(lambda: tf.group_by(
            ['group'], {'mean_val': (np.mean, 'value'), 'sum_val': (np.sum, 'value')}))
        t_pandas = median_of(
            lambda: df.groupby('group')['value'].agg(['mean', 'sum']).reset_index())
        t_polars = None
        if HAS_POLARS:
            plf = pl.DataFrame(tf.data)
            t_polars = median_of(lambda: plf.group_by('group').agg(
                pl.col('value').mean().alias('mean_val'),
                pl.col('value').sum().alias('sum_val'),
            ))
        print_row(label, t_accel, t_pure, t_pandas, t_polars)

    # multi-column
    for n_rows, label in [(100_000, "100k, 2 col, ~300 grp"), (1_000_000, "1M, 2 col, ~300 grp")]:
        tf = Tafra({
            'g1': rng.integers(0, 100, size=n_rows).astype('int64'),
            'g2': rng.choice(['x', 'y', 'z'], size=n_rows),
            'value': rng.standard_normal(n_rows),
        })
        df = pd.DataFrame(tf.data)
        t_accel, t_pure = _tafra_time(lambda: tf.group_by(
            ['g1', 'g2'], {'mean_val': (np.mean, 'value')}))
        t_pandas = median_of(
            lambda: df.groupby(['g1', 'g2'])['value'].agg('mean').reset_index())
        t_polars = None
        if HAS_POLARS:
            plf = pl.DataFrame(tf.data)
            t_polars = median_of(lambda: plf.group_by('g1', 'g2').agg(
                pl.col('value').mean().alias('mean_val')))
        print_row(label, t_accel, t_pure, t_pandas, t_polars)


def bench_transform():
    print("\n--- Transform (mean) ---")
    print_header()
    rng = np.random.default_rng(42)

    for n_rows, n_groups, label in [
        (10_000, 50, "10k rows, 50 grp"),
        (100_000, 100, "100k rows, 100 grp"),
        (1_000_000, 1000, "1M rows, 1k grp"),
    ]:
        tf = Tafra({
            'group': rng.integers(0, n_groups, size=n_rows).astype('int64'),
            'value': rng.standard_normal(n_rows),
        })
        df = pd.DataFrame(tf.data)

        t_accel, t_pure = _tafra_time(lambda: tf.transform(
            ['group'], {'mean_val': (np.mean, 'value')}))
        t_pandas = median_of(
            lambda: df.assign(mean_val=df.groupby('group')['value'].transform('mean')))
        t_polars = None
        if HAS_POLARS:
            plf = pl.DataFrame(tf.data)
            t_polars = median_of(lambda: plf.with_columns(
                pl.col('value').mean().over('group').alias('mean_val')))
        print_row(label, t_accel, t_pure, t_pandas, t_polars)


def bench_joins():
    print("\n--- Inner Join (equi) ---")
    print_header()
    rng = np.random.default_rng(42)

    for n, label in [(1000, "1k x 1k"), (5000, "5k x 5k"), (10000, "10k x 10k"),
                     (50000, "50k x 50k")]:
        left_tf = Tafra({
            'key': rng.integers(0, 100, size=n).astype('int64'),
            'left_val': rng.standard_normal(n),
        })
        right_tf = Tafra({
            'key': rng.integers(0, 100, size=n).astype('int64'),
            'right_val': rng.standard_normal(n),
        })
        left_df = pd.DataFrame(left_tf.data)
        right_df = pd.DataFrame(right_tf.data)

        t_accel, t_pure = _tafra_time(lambda: left_tf.inner_join(
            right_tf, [('key', 'key', '==')]))
        t_pandas = median_of(lambda: pd.merge(
            left_df, right_df, on='key', how='inner'))
        t_polars = None
        if HAS_POLARS:
            left_pl = pl.DataFrame(left_tf.data)
            right_pl = pl.DataFrame(right_tf.data)
            t_polars = median_of(lambda: left_pl.join(right_pl, on='key', how='inner'))
        print_row(label, t_accel, t_pure, t_pandas, t_polars)

    print("\n--- Left Join (equi) ---")
    print_header()
    for n, label in [(1000, "1k x 1k"), (5000, "5k x 5k"), (50000, "50k x 50k")]:
        left_tf = Tafra({
            'key': rng.integers(0, 100, size=n).astype('int64'),
            'left_val': rng.standard_normal(n),
        })
        right_tf = Tafra({
            'key': rng.integers(0, 100, size=n).astype('int64'),
            'right_val': rng.standard_normal(n),
        })
        left_df = pd.DataFrame(left_tf.data)
        right_df = pd.DataFrame(right_tf.data)

        t_accel, t_pure = _tafra_time(lambda: left_tf.left_join(
            right_tf, [('key', 'key', '==')]))
        t_pandas = median_of(lambda: pd.merge(
            left_df, right_df, on='key', how='left'))
        t_polars = None
        if HAS_POLARS:
            left_pl = pl.DataFrame(left_tf.data)
            right_pl = pl.DataFrame(right_tf.data)
            t_polars = median_of(lambda: left_pl.join(right_pl, on='key', how='left'))
        print_row(label, t_accel, t_pure, t_pandas, t_polars)


if __name__ == '__main__':
    print("=" * 80)
    libs = f"pandas {pd.__version__}, numpy {np.__version__}"
    if HAS_POLARS:
        libs += f", polars {pl.__version__}"
    accel_status = "C extension ACTIVE" if HAS_ACCEL else "C extension NOT available"
    print(f"Tafra Performance Comparison ({accel_status})")
    print(libs)
    print(f"All times in ms. * = fastest in row.")
    print("=" * 80)

    bench_construction()
    bench_column_access()
    bench_row_mapping()
    bench_groupby()
    bench_transform()
    bench_joins()

    print("\n" + "=" * 80)
    print("Done.")



"""
Tafra: a minimalist dataframe

Copyright (c) 2020 Derrick W. Turk and David S. Fulford

Author
------
Derrick W. Turk
David S. Fulford

Notes
-----
Created on April 25, 2020
"""
from typing import Callable, Dict, Tuple, Any, Iterator, MutableMapping, Type, Optional

import numpy as np


class ObjectFormatter(Dict[str, Callable[[np.ndarray[Any, Any]], np.ndarray[Any, Any]]],
                      MutableMapping[str, Callable[[np.ndarray[Any, Any]], np.ndarray[Any, Any]]]):
    """
    A dictionary that contains mappings for formatting objects. Some numpy objects
    should be cast to other types, e.g. the :class:`decimal.Decimal` type cannot
    operate with :class:`np.float`. These mappings are defined in this class.

    Each mapping must define a function that takes a :class:`np.ndarray` and
    returns a :class:`np.ndarray`.

    The key for each mapping is the name of the type of the actual value,
    looked up from the first element of the :class:`np.ndarray`, i.e.
    ``type(array[0]).__name__``.
    """
    test_array = np.arange(4)

    def __setitem__(self, dtype: str,
                    value: Callable[[np.ndarray[Any, Any]], np.ndarray[Any, Any]],
                    ) -> None:
        """
        Set the dtype formatter.
        """
        try:
            result = value(self.test_array)
        except Exception as e:
            raise ValueError(
                'Must provide a function that takes an ``np.ndarray`` and returns '
                'an np.ndarray.') from e

        if not isinstance(result, np.ndarray):
            raise ValueError(
                'Must provide a function that takes an ``np.ndarray`` and returns '
                'an np.ndarray.')

        dict.__setitem__(self, dtype, value)

    def __getitem__(self, dtype: str) -> Callable[[np.ndarray[Any, Any]], np.ndarray[Any, Any]]:
        """
        Get the dtype formatter.
        """
        return dict.__getitem__(self, dtype)

    def __delitem__(self, dtype: str) -> None:
        """
        Delete the dtype formatter.
        """
        dict.__delitem__(self, dtype)

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        if self.__len__() < 1:
            return r'{}'
        return '{' + '\n'.join(f'{c}: {v}' for c, v in self.items()) + '}'

    def __iter__(self) -> Iterator[Any]:
        yield from dict.__iter__(self)

    def __len__(self) -> int:
        return dict.__len__(self)

    def copy(self) -> Dict[str, Any]:
        return {k: dict.__getitem__(self, k) for k in self}

    def parse_dtype(self, value: np.ndarray[Any, Any]) -> Optional[np.ndarray[Any, Any]]:
        """
        Parse an object dtype.

        Parameters
        ----------
            value: np.ndarray
                The :class:`np.ndarray` to be parsed.

        Returns
        -------
            value, modified: Tuple(np.ndarray, bool)
                The :class:`np.ndarray` and whether it was modified or not.
        """
        if value.dtype.kind != 'O':
            return None

        type_name = type(value[0]).__name__
        if type_name in self:
            value = self[type_name](value)
            return value

        # convert object arrays of strings to StringDType
        if type_name == 'str':
            return value.astype(np.dtypes.StringDType())

        return None



"""Build configuration for optional C extension.

The C extension is optional — if compilation fails (missing compiler,
missing headers, etc.), the package installs without it and falls back
to pure Python + numpy at runtime.
"""
import os
import sys
import numpy as np
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext


class OptionalBuildExt(build_ext):
    """build_ext that treats C extension failure as non-fatal."""

    def build_extension(self, ext):
        try:
            super().build_extension(ext)
        except Exception as e:
            print(f"\n*** WARNING: Failed to build C extension: {e}")
            print("*** tafra will use pure Python + numpy (slower but functional).\n")


ext_modules = [
    Extension(
        "tafra._accel",
        sources=["tafra/_accel.c"],
        include_dirs=[np.get_include()],
    ),
]

setup(
    ext_modules=ext_modules,
    cmdclass={"build_ext": OptionalBuildExt},
)
